from grapevne.helpers import (
    init,
    script,
    resource,
    input,
    output,
    log,
    env,
    param,
    params,
)
from grapevne.helpers.helpers import (
    Helper,
    _helper,
)
from unittest import mock
from pathlib import Path
import pytest


class Workflow:
    def __init__(self, config):
        self.config = config


def test_script():
    init()
    with mock.patch(
        "grapevne.helpers.helpers.HelperBase._get_file_path",
        lambda self, path: Path("workflows") / path,
    ):
        assert script("script.py") == Path("workflows/scripts/script.py")


def test_resource():
    init()
    with mock.patch(
        "grapevne.helpers.helpers.HelperBase._get_file_path",
        lambda self, path: Path("workflows") / path,
    ):
        assert resource("resource.txt") == Path("workflows/../resources/resource.txt")


def test_get_port_spec_null():
    helper = Helper()
    port = None
    assert helper._get_port_spec(port) == []


def test_get_port_spec_str():
    helper = Helper()
    port = "incoming_namespace_1"
    expected_get_port_spec = [
        {
            "ref": "in",
            "label": "In",
            "namespace": "incoming_namespace_1",
        },
    ]
    assert helper._get_port_spec(port) == expected_get_port_spec


def test_get_port_spec_dict_shorthand():
    helper = Helper()
    port = {
        "port1": "incoming_namespace_1",
        "port2": "incoming_namespace_2",
    }
    expected_get_port_spec = [
        {
            "ref": "port1",
            "label": "port1",
            "namespace": "incoming_namespace_1",
        },
        {
            "ref": "port2",
            "label": "port2",
            "namespace": "incoming_namespace_2",
        },
    ]
    assert helper._get_port_spec(port) == expected_get_port_spec


def test_get_port_spec_dict():
    helper = Helper()
    port = {
        "ref": "port1",
        "label": "port1",
        "namespace": "incoming_namespace_1",
    }
    expected_get_port_spec = [
        {
            "ref": "port1",
            "label": "port1",
            "namespace": "incoming_namespace_1",
        },
    ]
    assert helper._get_port_spec(port) == expected_get_port_spec


def test_get_port_spec_list():
    helper = Helper()
    expected_get_port_spec = [
        {
            "ref": "port1",
            "label": "port1",
            "namespace": "incoming_namespace_1",
        },
        {
            "ref": "port2",
            "label": "port2",
            "namespace": "incoming_namespace_2",
        },
    ]
    assert helper._get_port_spec(expected_get_port_spec) == expected_get_port_spec


def test_input_single():
    workflow = Workflow(
        {
            "input_namespace": "in",
        }
    )
    init(workflow)
    assert Path(input("infile.txt")) == Path("results/in/infile.txt")


def test_input_multi():
    workflow = Workflow(
        {
            "input_namespace": {
                "port1": "in1",
                "port2": "in2",
            },
        }
    )
    init(workflow)
    assert Path(input("infile1.txt", "port1")) == Path("results/in1/infile1.txt")
    assert Path(input("infile2.txt", "port2")) == Path("results/in2/infile2.txt")


def test_get_namespace():
    workflow = Workflow(
        {
            "namespace": "namespace1",
        }
    )
    init(workflow)
    assert _helper._get_namespace() == "namespace1"


def test_get_namespace_none():
    workflow = Workflow(
        {
            "no_namespace_given": "namespace1",
        }
    )
    init(workflow)
    assert _helper._get_namespace() is None


def test_get_namespace_legacy():
    """Legacy 'output_namespace' check"""
    workflow = Workflow(
        {
            "output_namespace": "namespace1",
        }
    )
    init(workflow)
    assert _helper._get_namespace() == "namespace1"


def test_output():
    workflow = Workflow(
        {
            "output_namespace": "out",
        }
    )
    init(workflow)
    assert Path(output("outfile.txt")) == Path("results/out/outfile.txt")


def test_log():
    init()
    assert log("rule.log") == "logs/rule.log"


def test_env():
    init()
    assert env("conda.yaml") == "envs/conda.yaml"


def test_param():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
                "param2": {
                    "param3": "value3",
                },
            },
        }
    )
    init(workflow)
    # Single value
    assert params("param1") == "value1"
    # Long-form (comma-separated arguments)
    assert params("param2", "param3") == "value3"
    # Short-hand (dot-separated string)
    assert params("param2.param3") == "value3"


def test_param_notfound():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
            },
        }
    )
    init(workflow)
    with pytest.raises(ValueError):
        param("param2")


def test_param_default():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
            },
        }
    )
    init(workflow)
    default = "default_value"
    assert param("param2", default=default) == default


def test_params():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
                "param2": {
                    "param3": "value3",
                },
            },
        }
    )
    init(workflow)
    # Single value
    assert params("param1") == "value1"
    # Long-form (comma-separated arguments)
    assert params("param2", "param3") == "value3"
    # Short-hand (dot-separated string)
    assert params("param2.param3") == "value3"


def test_params_notfound():
    workflow = Workflow(
        {
            "params": {
                "param1": "value1",
            },
        }
    )
    init(workflow)
    with pytest.raises(ValueError):
        params("param2")



import os
import logging
import ensurepip
import subprocess

from urllib.parse import urlparse
from abc import ABC, abstractmethod
from pathlib import Path
from packaging.version import Version
from contextlib import contextmanager
from typing import Union
from grapevne.utils import (
    get_ports,
    get_port_spec,
    get_namespace,
    get_port_namespace,
)

try:
    import snakemake

except ImportError:
    raise ImportError(
        "This library must be called from within a snakemake environment."
    )

_expose_methods = [
    "script",
    "resource",
    "remote",
    "input",
    "output",
    "log",
    "env",
    "benchmark",
    "param",
    "params",
]


class HelperBase(ABC):
    """Abstract Base Class (ABC) for helper functions"""

    def __init__(self, workflow=None):
        self.workflow = workflow
        self.config = workflow.config if workflow else None

    def _check_config(self):
        if self.config is None:
            raise ValueError("Configuration not loaded")

    def _check_workflow(self):
        if self.workflow is None:
            raise ValueError("Workflow not loaded")

    # Utility functions

    @abstractmethod
    def _workflow_path(self, path):
        pass

    @abstractmethod
    def _get_remote_file_path(self, path, provider=None):
        pass

    def _get_provider(self, path):
        parsed = urlparse(path)
        # We need to account for Windows which returns 'C' for 'C:/
        if os.path.isabs(path) or os.path.exists(path) or not parsed.scheme:
            return ""
        return parsed.scheme

    def _get_file_path(self, path):
        path = self._workflow_path(str(path))
        provider = self._get_provider(path)
        if provider:
            return self._get_remote_file_path(path)
        return path

    def _module_path(self, base, path):
        """Return the path to a module file"""
        folder = f"{base}"
        module_name = self._get_namespace()
        folder += f"/{module_name}" if module_name else ""
        return f"{folder}/{path}"

    def _get_port_spec(self, port: Union[str, dict, list, None]):
        return get_port_spec(port)

    def _get_ports(self, config):
        return get_ports(config)

    def _get_port_namespace(self, ports, port_ref):
        return get_port_namespace(ports, port_ref)

    def _get_namespace(self):
        return get_namespace(self.config)

    # File-type specialisations

    def script(self, relpath):
        """Return the path to a script file"""
        return self._get_file_path(Path("scripts") / relpath)

    def resource(self, relpath):
        """Return the path to a resource file"""
        return self._get_file_path(Path("..") / "resources" / relpath)

    def remote(self, path):
        """Return the path to a remote file"""
        return self._get_remote_file_path(str(path))

    def input(self, path=None, port=None):
        """Return the path to an input file

        Args:
            path (str): The path to the input file
            port (str): The name of the input port (optional)
        """
        self._check_config()
        ports = self._get_ports(self.config)
        if len(ports) == 0:
            raise ValueError("No input ports defined in config")
        if port:  # port ref specified
            indir = self._get_port_namespace(ports, port)
        elif len(ports) == 1:
            indir = ports[0]["namespace"]
        else:
            raise Exception(f"Port not specified for input: {path} {port}")
        if path:
            return f"results/{indir}/{path}"
        else:
            return f"results/{indir}"

    def output(self, path=None):
        """Return the path to an output file"""
        self._check_config()
        outdir = self._get_namespace()
        if not outdir:
            raise ValueError("Output namespace not defined in config")
        if path:
            return f"results/{outdir}/{path}"
        else:
            return f"results/{outdir}"

    def log(self, path=None):
        """Return the path to a log file"""
        path = path if path else "log.txt"
        return self._module_path("logs", path)

    def env(self, path):
        """Return the path to an environment file"""
        return f"envs/{path}"

    def benchmark(self, path=None):
        """Return the path to a benchmark file"""
        path = path if path else "benchmark.tsv"
        return self._module_path("benchmarks", path)

    # Parameter indexing

    def params(self, *args, default=None):
        """Return the value of a parameter in the configuration

        Args:
            *args:  The path to the parameter in the configuration
                    Example: params("foo", "bar") will return the value of
                        config["params"]["foo"]["bar"]
                    You can also use the shorthand params("foo.bar"), so long as a
                    unique key "foo.bar" does not already exist.
                    If no arguments are provided, return the entire params dictionary
        """
        self._check_config()
        if len(args) == 0:
            return self.config.get("params", {})
        value = self.config.get("params", {})
        # Check if the argument is a dot-separated string
        if len(args) == 1:
            if args[0] not in value:
                args = args[0].split(".")
        # Iterate through the arguments
        for arg in args:
            if not isinstance(value, dict):
                raise ValueError(f"Attempting to index a non-indexable value: {value}")
            if arg not in value:
                if default is not None:
                    return default
                raise ValueError(f"Parameter not found: {arg}")
            value = value.get(arg, {})
        return value

    # Alias function - 'params' is the Snakefile directive, but 'param' is more
    # intuitive and consistent with other helper functions.
    def param(self, *args, **kwargs):
        return self.params(*args, **kwargs)


class HelperSnakemake7(HelperBase):
    """Helper class for Snakemake 7"""

    def __init__(self, workflow=None):
        super().__init__(workflow)

    # Implementations of abstract methods

    def _workflow_path(self, path):
        return snakemake.workflow.srcdir(path)

    def _get_remote_file_path(self, path, provider=None):
        from snakemake.remote import AUTO

        return next(iter(AUTO.remote(path)))


class HelperSnakemake8(HelperBase):
    """Helper class for Snakemake 8"""

    def __init__(self, workflow=None):
        super().__init__(workflow)

    # Utility functions

    def _install_remote_provider(self, path, provider=None):
        # Determine which storage plugin(s) to install
        if not provider:  # Infer provider if not explicit
            provider = self._get_provider(path)
        install_plugins = []
        if provider in ["http", "https"]:
            install_plugins += ["http"]
        if provider == "s3":
            install_plugins += ["s3"]
        if len(install_plugins) == 0:
            logging.warn(
                f"Provider '{provider}' not recognised - "
                "cannot install storage plugin(s)."
            )
            return
        # Install plugins and register with snakemake
        ensurepip.bootstrap()
        for plugin in install_plugins:
            subprocess.run(
                ["pip", "install", f"snakemake-storage-plugin-{plugin}"], check=True
            )
        # Register the plugins with snakemake
        from snakemake_interface_storage_plugins.registry import (
            StoragePluginRegistry,
        )

        StoragePluginRegistry().collect_plugins()

    # Implementations of abstract methods

    def _workflow_path(self, relpath):
        basedir = str(self.workflow.current_basedir)
        provider = self._get_provider(basedir)
        if provider:
            path = f"{basedir}/{relpath}"
        else:
            path = str(Path(basedir) / relpath)
        return path.replace("workflow/../", "")

    def _get_remote_file_path(self, path, provider=None):
        try:
            return self.workflow.storage_registry(path)
        except snakemake.exceptions.WorkflowError:
            logging.warn(
                "Error loading remote file - "
                "attempting to install provider and retrying."
            )
            self._install_remote_provider(path, provider)
            return self.workflow.storage_registry(path)


class Helper:
    """Wrapper class to handle different versions of snakemake"""

    def __new__(cls, workflow=None):
        snakemake_version = Version(snakemake.__version__)
        if snakemake_version < Version("7"):
            raise ValueError("GRAPEVNE requires snakemake version 7 or higher")
        if snakemake_version < Version("8"):
            return HelperSnakemake7(workflow)
        else:
            return HelperSnakemake8(workflow)


@contextmanager
def grapevne_helper(globals_dict):
    workflow = globals_dict.get("workflow", None)
    gv = Helper(workflow)

    # Expose methods from Helper as method in calling (globals) namespace
    for name in _expose_methods:
        globals_dict[name] = getattr(gv, name)

    try:
        # Yield the Helper object as a context manager
        yield gv
    finally:
        # Clean up the globals namespace
        for name in _expose_methods:
            del globals_dict[name]


_helper = Helper()


def init(workflow=None):
    _helper.workflow = workflow
    _helper.config = workflow.config if workflow else None



from typing import List, Union, TypedDict


Namespace = Union[list, None, str, dict]


class SnakefileRemotePath(TypedDict):
    function: str
    args: List[str]
    kwargs: dict


Snakefile = Union[str, SnakefileRemotePath]


# Input port specification
class Port:
    def __init__(
        self,
        ref: str,
        label: str,
        namespace: str,
        mapping: list[dict[str, str]]
    ):
        """Initialise a Port object

        Args:
            ref (str): Port reference (used in Snakefile)
            label (str): Port label (displayed in UI)
            namespace (str): Incoming / input namespace
            mapping (list[dict[str, str]]): Nested structures {module: (str), port: (str)}
        """
        self.ref = ref
        self.label = label
        self.namespace = namespace
        self.mapping = mapping


class Module:
    def __init__(
        self,
        name: str,
        rulename: str,
        nodetype: str,
        snakefile: Snakefile = "",
        config=None,
        ports: [Port] = None,
        namespace: str = "",
        docstring: str = "",  # passthrough (unused in builds)
        # Backwards compatiblity list
        input_namespace: Namespace = None,
        output_namespace: Union[str, None] = None,
    ):
        """Initialise a Node object, the parent class for Modules

        Args:
            name (str): Name of the node
            rulename (str): Name of the rule
            nodetype (str): Type of node (module, connector, etc.)
            snakefile (str|dict): str location or dict representing function call
            config (dict): Configuration (parameters) for the Snakefile
            ports ([Port]): List of input ports
            namespace (str): Module namespace
        """

        self.name = name
        self.rulename = rulename
        self.nodetype = nodetype
        self.snakefile = snakefile
        self.config = {} if not config else config
        self.ports = [] if not ports else ports
        self.namespace = namespace

        # Backwards compatibility (input_namespace -> ports)
        if ports is not None and input_namespace is not None:
            raise ValueError(
                "Cannot specify both 'ports' and (legacy) 'input_namespace' in Module"
            )
        if input_namespace is not None:
            self.ports = get_port_spec(input_namespace)
        if "input_namespace" in self.config.keys():
            del self.config["input_namespace"]

        # Backwards compatibility (output_namespace -> namespace)
        if namespace is not None and output_namespace is not None:
            raise ValueError(
                "Cannot specify both 'namespace' and (legacy) 'output_namespace' in Module"
            )
        if output_namespace is not None:
            self.namespace = output_namespace
        if "output_namespace" in self.config.keys():
            del self.config["output_namespace"]

        if not isinstance(self.ports, list):
            raise ValueError(f"({self.name}): Ports must be a list, not {type(self.ports)}")


def get_port_spec(port: Union[str, dict, list, None]) -> Port:
    """Input port specification

    Utility function to convert shorthand input port specifications to their full
    format. Always returns a list.

    Port specification:
        [
            {
                ref: (required; str),       # Port reference (used in Snakefile)
                label: (required; str),
                namespace: (required; str), # Namespace link (incoming)

                # Used for composite modules:
                mapping: [
                    {                       #   Target module
                        module: (str),      #     target module name / reference
                        port: (str),        #     target port reference

                    },
                    ...
                ],
            },
            ...
        ]

    Shorthand specifications:
        null    No input ports
        str     Single input port
        dict    Multiple or named input ports, with port names as keys and
                namespaces as values
    """

    if port is None:
        # No input ports (null)
        return []
    if isinstance(port, str):
        # Single input port (str)  - convert to dictionary for conversion
        port = {"in": port}
    if isinstance(port, dict):
        # First, check that the dictionary is not a new-form Port definition
        required_keys = ["ref", "label", "namespace"]
        if all(key in port for key in required_keys):
            # single dict (new format)
            return [port]
        new_ports = []
        for k, v in port.items():
            p = {
                "ref": k,
                "label": "In" if k == "in" else k,
                "namespace": v,
            }
            # Check for old-form node mapping and convert to new form
            if "$" in k:
                module, port = k.split('$')  # Only map one-connection deep
                p['mapping'] = [{
                    'module': module,
                    'port': port if port else "in",
                }]
            # Append port
            new_ports.append(p)
        return new_ports
    if isinstance(port, list):
        return port
    raise ValueError(f"Unknown port specification: {port}")


def get_ports(config):
    """Get ports specification from config

    Includes a backwards compatibility check for the legacy 'input_namespace'
    """
    ports = config.get("ports", config.get("input_namespace", None))
    return get_port_spec(ports)


def get_port_namespace(ports, port_ref=None):
    """Get the namespace for a given port reference"""
    ports = get_port_spec(ports)
    if port_ref is None:
        if len(ports) == 0:
            return []
        if len(ports) == 1:
            return ports[0]["namespace"]
        else:
            raise ValueError("Port reference required when multiple ports are present (ports: {ports})")
    for port in ports:
        if port["ref"] == port_ref:
            return port["namespace"]
    raise ValueError(f"Port not found: {port_ref}")


def get_namespace(config):
    """Return the module's namespace

    Includes a backwards compatibility check for the legacy 'output_namespace'
    """
    namespace = None
    if config:
        namespace = config.get("namespace", None)
        if namespace is None:
            # Backwards compatibility
            namespace = config.get("output_namespace", None)
    return namespace
